/*------------------------------------------------------------------------
  File        : rtb_rpt_class.p
  Description : Class-specific Report Library
  ----------------------------------------------------------------------*/
USING rtb.repository.* FROM PROPATH.

{rtb/i/rtb_xreftype.i}
{rtb/i/rtb_rptinit.i}
{rtb/i/rtb_attribtypes.i}

DEFINE TEMP-TABLE ttClass
  FIELD id       AS INTEGER    /* record id */
  FIELD refType  AS INTEGER    /* type of reference */
  FIELD ref      AS CHARACTER  /* reference text */
  FIELD misc     AS CHARACTER  /* extra ref info (desc, params, etc */
  INDEX index1 IS PRIMARY UNIQUE id
  INDEX index2 refType id.

/*
 These preprocessors are used as report sections for the Class Details report and as
 a means to sort output.  Changing a number will change the section's order in the
 report output.
*/
&SCOPED-DEFINE SUPERCLASS 1
&SCOPED-DEFINE SUBCLASS 3
&SCOPED-DEFINE INTERFACE 2
&SCOPED-DEFINE METHOD 4
&SCOPED-DEFINE DATAMEMBER 5
&SCOPED-DEFINE PROPERTY 6

PROCEDURE add_reference PRIVATE :
/*------------------------------------------------------------------------------
  Purpose:    Add records to the ttClass temp-table
  Parameters: INPUT pcRefType  - the type of reference (see preprocessors)
              INPUT pcRefText  - referenced object
              INPUT pcMisc     - extra info about the object
  Notes:      
------------------------------------------------------------------------------*/
  DEFINE INPUT PARAMETER pcRefType  AS INTEGER  NO-UNDO.
  DEFINE INPUT PARAMETER pcRefText  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER pcMisc     AS CHARACTER NO-UNDO.

  DEFINE VARIABLE iRefCnt AS INTEGER NO-UNDO.

  DEFINE BUFFER bttClass FOR ttClass.

  FIND LAST bttClass NO-ERROR.
  IF AVAILABLE bttClass THEN 
    iRefCnt = bttClass.id + 1.
  ELSE
    iRefCnt = 1.

  CREATE ttClass.
  ASSIGN
    ttClass.id       = iRefCnt
    ttClass.refType  = pcRefType
    ttClass.ref      = pcRefText
    ttClass.misc     = pcMisc.

  RETURN.

END PROCEDURE.

PROCEDURE find_other_references PRIVATE :
/*------------------------------------------------------------------------------
  Purpose:     Find other references we might be interested in.
  Parameters:  pcObjGuid
  Notes:       This looks at 'xref data' for the given class for interface
               implementations, data-members, and methods
------------------------------------------------------------------------------*/
  DEFINE INPUT  PARAMETER pcObjGuid AS CHARACTER NO-UNDO.

  DEFINE VARIABLE cAttributes AS CHARACTER NO-UNDO INITIAL "".
  
  DEFINE BUFFER Brtb_object  FOR rtb.rtb_object.
  DEFINE BUFFER Brtb_ver     FOR rtb.rtb_ver.
  DEFINE BUFFER Brtb_xref    FOR rtb.rtb_xref.

  FOR EACH Brtb_xref NO-LOCK
    WHERE Brtb_xref.src-object-guid = pcObjGuid:

    /*
     Interface implementations.. 
    */
    IF Brtb_xref.ref-type = {&XREF_IMPLEMENTATION} THEN DO:

      /*
       Could be an implmentation of a interface no in workspace (i.e. OE class)
      */
      IF Brtb_xref.ref-object-guid-1 = "" THEN
        NEXT.

      FIND Brtb_object NO-LOCK
        WHERE Brtb_object.object-guid = Brtb_xref.ref-object-guid-1.
        
      FIND Brtb_ver NO-LOCK OF Brtb_object.

      RUN add_reference ({&INTERFACE},Brtb_xref.ref-text,Brtb_ver.description).
  
    END.
    
    /*
     In the code below, we only shows methods, properties, and data-members that have attributes
     associated with them. This will ensure that we only show these items that are defined in the class
     for which we are reporting. Previously we checked the ref-actions but that was not always accurate
     as it filtered out internal method references if they happened to also be called internally.
    */

    /*
     Defined methods..
    */
    IF Brtb_xref.ref-type = {&XREF_METHOD} THEN DO:

      cAttributes = AttributeHandler:getAttributes(Brtb_xref.xref-guid, {&ATTRIB-XREF}).       
        
      IF cAttributes <> "" THEN
        RUN add_reference ({&METHOD},Brtb_xref.obj-text, cAttributes).
        
      cAttributes = "".
        
    END.  

    /*
     Data members.. 
    */
    IF Brtb_xref.ref-type = {&XREF_DATAMEMBER} THEN DO:
      
      cAttributes = AttributeHandler:getAttributes(Brtb_xref.xref-guid, {&ATTRIB-XREF}).
      
      IF cAttributes <> "" THEN 
        RUN add_reference ({&DATAMEMBER},Brtb_xref.obj-text, AttributeHandler:getAttributes(Brtb_xref.xref-guid, {&ATTRIB-XREF})).
        
      cAttributes = "".
      
    END.

    /*
     Properties.. 
    */
    IF Brtb_xref.ref-type = {&XREF_PROPERTY} THEN DO:
      
      cAttributes = AttributeHandler:getAttributes(Brtb_xref.xref-guid, {&ATTRIB-XREF}).
      
      IF cAttributes <> "" THEN       
        RUN add_reference ({&PROPERTY},Brtb_xref.obj-text, AttributeHandler:getAttributes(Brtb_xref.xref-guid, {&ATTRIB-XREF})).
        
      cAttributes = "".
      
    END.

  END.

  RETURN.

END PROCEDURE.

PROCEDURE find_subclasses PRIVATE:
/*------------------------------------------------------------------------------
  Purpose:     Find any subclasses of the given class.
  Parameters:  pcObjGuid
  Notes:       This looks at 'where-used' data to find if any other class has
               subclassed this one.
------------------------------------------------------------------------------*/
  DEFINE INPUT PARAMETER pcObjGuid AS CHARACTER NO-UNDO.

  DEFINE BUFFER Brtb_object  FOR rtb.rtb_object.
  DEFINE BUFFER Brtb_ver     FOR rtb.rtb_ver.
  DEFINE BUFFER Brtb_xref    FOR rtb.rtb_xref.

  FOR EACH Brtb_xref NO-LOCK
    WHERE Brtb_xref.ref-object-guid-1 = pcObjGuid
      AND Brtb_xref.src-type   = {&XREF_CLASS}:

    IF Brtb_xref.ref-type = {&XREF_SUBCLASS} THEN DO:

      FIND Brtb_object NO-LOCK
        WHERE Brtb_object.object-guid = Brtb_xref.src-object-guid.

      FIND Brtb_ver NO-LOCK OF Brtb_object.

      RUN add_reference ({&SUBCLASS},Brtb_object.object,Brtb_ver.description).

    END.

  END.

  RETURN.
  
END PROCEDURE.

PROCEDURE find_superclasses PRIVATE:
/*------------------------------------------------------------------------------
  Purpose:     Look for superClass records in the class hierarchy for the given
               class and add them to our temp-table.
  Parameters:  pcObjGuid
  Notes:       We do a little recursion here to build the list of superClasses
------------------------------------------------------------------------------*/
  DEFINE INPUT PARAMETER pcObjGuid AS CHARACTER NO-UNDO.

  DEFINE BUFFER Brtb_object FOR rtb.rtb_object.
  DEFINE BUFFER Brtb_xref   FOR rtb.rtb_xref.

  FIND Brtb_object NO-LOCK
    WHERE Brtb_object.object-guid = pcObjGuid.

  RUN add_reference ({&SUPERCLASS},Brtb_object.object,"").

  /*
   Find superClass of class
  */
  FIND FIRST Brtb_xref NO-LOCK
    WHERE Brtb_xref.src-object-guid = Brtb_object.object-guid
      AND Brtb_xref.ref-type  = {&XREF_SUBCLASS} NO-ERROR.

  /*
   If ref-object-guid-1 is not blank, then the superclass is an object known to
   Roundtable. Otherwise, just simply add the obj-text which will contain
   the name of the parent class.
  */
  IF AVAILABLE Brtb_xref THEN DO:
    IF Brtb_xref.ref-object-guid-1 <> "" THEN
      RUN find_superclasses (Brtb_xref.ref-object-guid-1). /* ooh.. ah.. recursion */
    ELSE IF Brtb_xref.obj-text <> "" THEN
      RUN add_reference ({&SUPERCLASS}, Brtb_xref.obj-text, "").
  END.

  RETURN.

END PROCEDURE.

PROCEDURE rtb_class_details PRIVATE:
/*------------------------------------------------------------------------------
  Purpose:    Main procedure for the Class Details report
  Parameters: INPUT Pwspace  - workspace id
              INPUT PobjType - object type
              INPUT Pobject  - class to report on
  Notes:       
------------------------------------------------------------------------------*/
  DEFINE INPUT PARAMETER Pwspace  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER PobjType AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER Pobject  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER Ppmod    AS CHARACTER NO-UNDO.

  DEFINE VARIABLE i AS INTEGER NO-UNDO.

  DEFINE BUFFER Brtb_object FOR rtb.rtb_object.

  IF LOOKUP(PobjType,"PDASE,PFILE,PFIELD") <> 0 THEN
    FIND Brtb_object NO-LOCK
      WHERE Brtb_object.wspace-id = Pwspace
        AND Brtb_object.obj-type  = PobjType
        AND Brtb_object.object    = Pobject NO-ERROR.
  ELSE 
    FIND Brtb_object NO-LOCK
      WHERE Brtb_object.wspace-id = Pwspace
        AND Brtb_object.obj-type  = PobjType
        AND Brtb_object.object    = Pobject 
        AND Brtb_object.pmod      = Ppmod NO-ERROR.

  IF NOT AVAILABLE Brtb_object THEN
    RETURN "Object record is not available.".

  /*
   Gather report data..
  */
  RUN find_superclasses (Brtb_object.object-guid).
  RUN find_subclasses (Brtb_object.object-guid).  
  RUN find_other_references (Brtb_object.object-guid).

  /*
   Print the report.. 
  */
  {rtb/i/rtb_rpthdr.i 
     &title="'Class Details'"
     &criteria="'Class: ' Brtb_object.object"
     &center-criteria=NO
     &STREAM=Sout}

  /*
   superClass hiearchy..
  */
  PUT STREAM sOut "Class Hierarchy:" SKIP(1).

  FOR EACH ttClass
    WHERE ttClass.refType = {&SUPERCLASS}
    BY ttClass.id DESCENDING:

    i = i + 1.

    PUT STREAM sOut UNFORMATTED FILL(" ", i * 4 - 2). /* this faaancy calculation indents two spaces for each branch */

    IF i GT 1 THEN
        PUT STREAM sOut UNFORMATTED "|_". /* we only want this on secondary levels */

    PUT STREAM sOut UNFORMATTED ttClass.ref.

    IF ttClass.ref = Brtb_object.object THEN
      PUT STREAM sOut UNFORMATTED " *".

    PUT STREAM sOut SKIP.

  END.

  /*
   Process the rest of the temp-table breaking at each section.
  */
  FOR EACH ttClass
    BREAK BY ttClass.refType:

    IF FIRST-OF(ttClass.refType) THEN DO:

      CASE ttClass.refType:

      WHEN {&INTERFACE} THEN
        PUT STREAM sOut SKIP(2) "All Implemented Interfaces:" SKIP.

      WHEN {&SUBCLASS} THEN
        PUT STREAM sOut SKIP(2) "Direct Known Subclasses:" SKIP.

      WHEN {&METHOD} THEN
        PUT STREAM sOut SKIP(2) "Method Summary:" SKIP.

      WHEN {&DATAMEMBER} THEN
        PUT STREAM sOut SKIP(2) "Data Members:" SKIP.

      WHEN {&PROPERTY} THEN
        PUT STREAM sOut SKIP(2) "Properties:" SKIP.

      END CASE.

    END. /* first-of */

    CASE ttClass.refType:

    WHEN {&INTERFACE} THEN
      DISPLAY STREAM sOut
        SPACE(2)
        ttClass.ref  FORMAT "x(40)" LABEL "Interface"
        ttClass.misc FORMAT "x(40)" LABEL "Description"
        WITH FRAME frInterface WIDTH 120 DOWN STREAM-IO.

    WHEN {&SUBCLASS} THEN
      DISPLAY STREAM sOut
        SPACE(2)
        ttClass.ref  FORMAT "x(40)" LABEL "Subclass"
        ttClass.misc FORMAT "x(40)" LABEL "Description"
        WITH FRAME frSubclass WIDTH 120 DOWN STREAM-IO.

    WHEN {&METHOD} THEN 
      DISPLAY STREAM sOut
        SPACE(2)
        ttClass.ref  FORMAT "x(40)" LABEL "Method"
        ttClass.misc FORMAT "x(40)" LABEL "Attributes"
        WITH FRAME frMethod WIDTH 120 DOWN STREAM-IO.

    WHEN {&DATAMEMBER} THEN
      DISPLAY STREAM sOut
        SPACE(2)
        ttClass.ref  FORMAT "x(40)" LABEL "Data Member"
        ttClass.misc FORMAT "x(40)" LABEL "Attributes"
        WITH FRAME frDatamember WIDTH 120 DOWN STREAM-IO.

    WHEN {&PROPERTY} THEN
      DISPLAY STREAM sOut
        SPACE(2)
        ttClass.ref  FORMAT "x(40)" LABEL "Property"
        ttClass.misc FORMAT "x(40)" LABEL "Attributes"
        WITH FRAME frProperty WIDTH 120 DOWN STREAM-IO.

    END CASE.

  END. /* each ttClass */

  {rtb/i/rtb_rptftr.i &STREAM=Sout}

  EMPTY TEMP-TABLE ttClass.
  
  RETURN.

END PROCEDURE.

PROCEDURE rptRtbClassDetails :
/*------------------------------------------------------------------------------
  Purpose:     
  Parameters:  <none>
  Notes:       
------------------------------------------------------------------------------*/
  DEFINE INPUT PARAMETER Pwspace  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER PobjType AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER Pobject  AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER Ppmod    AS CHARACTER NO-UNDO.
  DEFINE INPUT PARAMETER TABLE-HANDLE PhTT.

  RUN rtb_report_initialize (INPUT TABLE-HANDLE PhTT).

  RUN rtb_class_details (Pwspace, PobjType, Pobject, Ppmod).

  OUTPUT STREAM Sout CLOSE.

  RETURN RETURN-VALUE.

END PROCEDURE.
